import json
import os
import sys
import time
from enum import Enum
from json import JSONEncoder
from typing import Any, Callable

import requests
from requests.structures import CaseInsensitiveDict

from entegywrapper.errors import EntegyInvalidAPIKeyError, EntegyNoDataError

sys.path.append(os.path.dirname(__file__))

API_ENDPOINTS = {
    "AU": "https://api.entegy.com.au",
    "US": "https://api-us.entegy.com.au",
    "EU": "https://api-eu.entegy.com.au",
}


class EnumJSONEncoder(JSONEncoder):
    """
    Custom JSON encoder that handles ENUMs.
    """

    def default(self, obj):
        if isinstance(obj, Enum):
            return obj.value
        return super().default(obj)


class EntegyAPI:
    from .attendance_tracking.attendance_tracking import add_check_in, get_attended, get_attendees
    from .content.categories import (
        available_categories,
        create_categories,
        create_child_categories,
        delete_categories,
        deselect_categories,
        select_categories,
        update_category,
    )
    from .content.content import (
        add_children_content,
        create_content,
        delete_content,
        get_content,
        get_schedule_content,
        update_content,
    )
    from .content.documents import add_documents, add_external_content_documents
    from .content.multi_link import (
        add_multi_links,
        get_multi_links,
        remove_all_multi_links,
        remove_multi_link,
    )
    from .notification.notification import send_bulk_notification, send_notification
    from .plugins.ext_auth import external_authentication
    from .points.point_management import award_points, get_point_leaderboard, get_points
    from .profiles.profile_custom import (
        all_profile_custom,
        create_profile_custom,
        delete_profile_custom,
        get_profile_custom,
        update_profile_custom,
    )
    from .profiles.profile_links import (
        clear_profile_links,
        deselect_profile_links,
        multi_select_profile_links,
        page_profile_links,
        select_profile_link,
        selected_profile_links,
    )
    from .profiles.profile_payments import add_profile_payment
    from .profiles.profile_types import (
        all_profile_types,
        create_profile_type,
        delete_profile_type,
        get_profile_type,
        update_profile_type,
    )
    from .profiles.profiles import (
        all_profiles,
        create_profile,
        delete_profile,
        get_profile,
        send_welcome_email,
        sync_profile_block,
        sync_profiles,
        update_profile,
    )

    def __init__(
        self,
        api_key: str | list[str],
        api_secret: str | list[str],
        project_id: str,
        region: str = "AU",
    ):
        """
        Creates an Entegy API wrapper to interact with the specified project.

        Parameters
        ----------
            `api_key` (`str | list[str]`): Entegy API key(s)

            `api_secret` (`str | list[str]`): Entegy API secret key(s)

            `project_id` (`str`): Entegy project ID

            `region` (`str`, optional): project region: one of "AU", "US", "EU"; defaults to "AU"
        """
        if isinstance(api_key, list):
            assert isinstance(api_secret, list)
            assert len(api_key) == len(api_secret)
            assert all(isinstance(key, str) for key in api_key)
            assert all(isinstance(secret, str) for secret in api_secret)
        else:
            assert isinstance(api_key, str)
            assert isinstance(api_secret, str)
        assert isinstance(project_id, str)
        assert isinstance(region, str)
        assert region in API_ENDPOINTS.keys()

        if isinstance(api_key, list) and isinstance(api_secret, list):
            self.api_key = list(map(lambda x: x.strip(), api_key))
            self.api_secret = list(map(lambda x: x.strip(), api_secret))
        elif isinstance(api_key, list) or isinstance(api_secret, list):
            raise ValueError("API key and secret must both be lists or both be strings")
        else:
            self.api_key = api_key.strip()
            self.api_secret = api_secret.strip()

        self.current_key_pair = 0
        self.project_id = project_id.strip()

        self.headers = CaseInsensitiveDict()
        self.headers["Content-Type"] = "application/json"
        self.get_key()

        self.api_endpoint = API_ENDPOINTS[region]

    def get_key(self) -> str:
        """
        Returns the API Key. If a list of keys was provided, the current key is
        returned.

        Returns
        -------
            `str`: API Key
        """
        if isinstance(self.api_key, list):
            self.headers["Authorization"] = f"ApiKey {self.api_secret[self.current_key_pair]}"
            return self.api_key[self.current_key_pair]

        self.headers["Authorization"] = f"ApiKey {self.api_secret}"
        return self.api_key

    def cycle_key(self):
        """
        Cycles to the next API keypair, wrapping to the first where necessary.
        """
        self.current_key_pair = (self.current_key_pair + 1) % len(self.api_key)

    def get_endpoint(self) -> str:
        """
        Returns the endpoint URL.

        Returns
        -------
            `str`: API endpoint URL
        """
        return self.api_endpoint

    def request(self, method: Callable, endpoint: str, data: dict) -> dict:
        """
        Sends the given data to the given endpoint of the Entegy API, using the given
        method. Internalised to allow for automatic key cycling and error handling.

        Parameters
        ----------
            `method` (`Callable`): method to use to send the request

            `endpoint` (`str`): API endpoint to which to post

            `data` (`dict`): data to post

        Raises
        ------
            `EntegyInvalidAPIKeyError`: if the API keys are invalid

        Returns
        -------
            `dict`: response data
        """
        keys_attempted = 0
        failed_requests = 0

        data |= {"apiKey": self.get_key(), "projectId": self.project_id}

        response = None
        while response is None:
            response = method(
                endpoint, headers=self.headers, data=json.dumps(data, cls=EnumJSONEncoder)
            )

            try:
                response = response.json()
            except:
                failed_requests += 1
                if failed_requests >= 5:
                    raise EntegyNoDataError()

                response = None
                continue

            match response["response"]:
                case 403:  # invalid API key
                    failed_requests += 1
                    if failed_requests >= 5:
                        raise EntegyInvalidAPIKeyError()

                    response = None
                case 489:  # rate-limit
                    if keys_attempted >= len(self.api_key):
                        duration = response["resetDuration"]
                        print(f"Rate limit reached, waiting {duration} seconds")
                        time.sleep(duration + 2)
                        print("Continuing...")
                        keys_attempted = 0
                        response = None
                    else:
                        self.cycle_key()
                        key = self.get_key()
                        data["apiKey"] = key
                        print(f"Rate limit reached, trying alternate key: {key}")
                        keys_attempted += 1
                        response = None

        return response

    def post(self, endpoint: str, data: dict) -> dict:
        """
        Posts the given data to the given endpoint of the Entegy API.

        Parameters
        ----------
            `endpoint` (`str`): API endpoint to which to post

            `data` (`dict`): data to post

        Returns
        -------
            `dict`: response data
        """
        return self.request(requests.post, endpoint, data)

    def delete(self, endpoint: str, data: dict) -> dict:
        """
        Deletes the given data from the given endpoint of the Entegy API.

        Parameters
        ----------
            `endpoint` (`str`): API endpoint from which to delete

            `data` (`dict`): data to delete

        Returns
        -------
            `dict`: response data
        """
        return self.request(requests.delete, endpoint, data)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from entegywrapper.errors import EntegyFailedRequestError
from entegywrapper.schemas.content import Category, ContentCreate, TemplateType

if TYPE_CHECKING:
    from entegywrapper import EntegyAPI


def available_categories(
    self: EntegyAPI,
    template_type: TemplateType,
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
) -> list[Category]:
    """
    Returns the list of available categories for the page specified by the given
    identifier.

    Parameters
    ----------
        `template_type` (`TemplateType`): the templateType of the page

        `module_id` (`int`, optional): the moduleId of the page; defaults to `None`

        `external_reference` (`str`, optional): the externalReference of the page; defaults to `None`

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails

    Returns
    -------
        `list[Category]`: the available categories
    """
    data: dict[str, Any] = {"templateType": template_type}

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/Available", data=data)

    match response["response"]:
        case 200:
            return [Category(**category) for category in response["availableCategories"]]
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 402:
            raise EntegyFailedRequestError("templateType doesn't support categories")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )


def select_categories(
    self: EntegyAPI,
    template_type: TemplateType,
    categories: list[dict[str, Any]],
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
) -> bool:
    """
    Selects the specified categories for the specified content page.

    Parameters
    ----------
        `template_type` (`TemplateType`): the templateType of the page selecting the categories

        `categories` (`list[Category]`): the categories to select

        `module_id` (`int`, optional): the moduleId of the page selecting the categories; defaults to `None`

        `external_reference` (`str`, optional): the externalReference of the page selecting the categories; defaults to `None`

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails

    Returns
    -------
        `bool`: whether the categories were selected successfully
    """
    data = {
        "templateType": template_type,
        "categories": categories,
    }

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/Select", data=data)

    match response["response"]:
        case 200:
            return True
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 402:
            raise EntegyFailedRequestError("templateType doesn't support categories")
        case 404:
            raise EntegyFailedRequestError("No categories")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )


def deselect_categories(
    self: EntegyAPI,
    template_type: TemplateType,
    categories: list[Category],
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
) -> bool:
    """
    Deselects the specified categories for the specified content page.

    Parameters
    ----------
        `template_type` (`TemplateType`): the templateType of the page to unselect the categories from

        `categories` (`list[Category]`): the categories to select

        `module_id` (`int`, optional): the moduleId of the page to unselect the categories from; defaults to `None`

        `external_reference` (`str`, optional): the externalReference of the page to unselect the categories from; defaults to `None`

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails

    Returns
    -------
        `bool`: whether the categories were deselected successfully
    """
    data = {"templateType": template_type, "categories": categories}

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/Deselect", data=data)

    match response["response"]:
        case 200:
            return True
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 402:
            raise EntegyFailedRequestError("templateType doesn't support categories")
        case 404:
            raise EntegyFailedRequestError("Category doesn't exist")
        case 405:
            raise EntegyFailedRequestError("Category not selected")
        case 406:
            raise EntegyFailedRequestError("No categories to unselect")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )

    return False


def create_categories(
    self: EntegyAPI,
    template_type: TemplateType,
    categories: list[Category] | list[ContentCreate],
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
):
    """
    Creates categories under a root page.

    Parameters
    ----------
        `template_type` (`TemplateType`): the templateType of the page holding the categories

        `categories` (`list[Category | ContentCreate]]`): the categories to create

        `module_id` (`int`, optional): the moduleId of the page holding the categories; defaults to `None`

        `external_reference` (`str`, optional): the externalReference of the page holding the categories; defaults to `None`

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails
    """
    data = {
        "templateType": template_type,
        "categories": [category.model_dump() for category in categories],
    }

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/Create", data=data)

    match response["response"]:
        case 200:
            return
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 402:
            raise EntegyFailedRequestError("Can't create categories here")
        case 404:
            raise EntegyFailedRequestError("Duplicate externalReference")
        case 405:
            raise EntegyFailedRequestError("Missing name")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )


def create_child_categories(
    self: EntegyAPI,
    categories: list[Category],
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
):
    """
    Creates categories under another category.

    Parameters
    ----------
        `categories` (`list[Category]`): the categories to create

        `module_id` (`int`): the moduleId of the page holding the categories

        `external_reference` (`int`): the externalReference of the page holding the categories

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails
    """
    data = {"categories": [category.model_dump() for category in categories]}

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/CreateChild", data=data)

    match response["response"]:
        case 200:
            return
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 404:
            raise EntegyFailedRequestError("Duplicate externalReference")
        case 405:
            raise EntegyFailedRequestError("Missing name")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )


def update_category(
    self: EntegyAPI,
    name: str,
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
):
    """
    Changes the name of a category.

    Parameters
    ----------
        `name` (`str`): the new name of the category

        `module_id` (`int`, optional): the moduleId of the category; defaults to `None`

        `external_reference` (`str`, optional): the externalReference of the category; defaults to `None`

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails
    """
    data: dict[str, Any] = {"name": name}

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/Update", data=data)

    match response["response"]:
        case 200:
            return
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 402:
            raise EntegyFailedRequestError("Missing name")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )


def delete_categories(
    self: EntegyAPI,
    template_type: TemplateType,
    categories: list[Category],
    *,
    module_id: Optional[int] = None,
    external_reference: Optional[str] = None,
):
    """
    Creates categories under another category.

    Parameters
    ----------
        `template_type` (`TemplateType`): the templateType of the page

        `categories` (`list[Category]`): the categories to delete

        `module_id` (`int`, optional): the moduleId of the page; defaults to `None`

        `external_reference` (`str`, optional): the externalReference of the page; defaults to `None`

    Raises
    ------
        `ValueError`: if no identifier is specified

        `EntegyFailedRequestError`: if the API request fails
    """
    data = {
        "templateType": template_type,
        "categories": [category.model_dump() for category in categories],
    }

    if module_id is not None:
        data["moduleId"] = module_id
    elif external_reference is not None:
        data["externalReference"] = external_reference
    else:
        raise ValueError("Please specify an identifier")

    response = self.post(self.api_endpoint + "/v2/Categories/Delete", data=data)

    match response["response"]:
        case 200:
            return
        case 401:
            raise EntegyFailedRequestError("Missing Id")
        case 404:
            raise EntegyFailedRequestError("Category doesn't exist")
        case _:
            raise EntegyFailedRequestError(
                f"{response['response']}: {response.get('message', 'Unknown error')}"
            )



from __future__ import annotations

from enum import Enum, IntEnum
from typing import Any, Optional

from pydantic import BaseModel

from .page_settings import PageSetting


class TemplateType(Enum):
    @classmethod
    def _missing_(cls, value):
        """Enables case-insensitive lookup."""
        if value is None:
            raise ValueError("None is not a valid value for this Enum.")

        if not isinstance(value, str):
            raise TypeError(f"Expected string type for value, got {type(value).__name__} instead.")

        value = value.lower()
        for member in cls.__members__.values():
            if member.value.lower() == value:
                return member

    ABOUT = "About"
    ABSTRACT = "Abstract"
    ABSTRACTS = "Abstracts"
    EXHIBITOR = "Exhibitor"
    EXHIBITORS = "Exhibitors"
    FLOOR_PLAN = "FloorPlan"
    FLOOR_PLANS = "FloorPlans"  # Doesn't exist in API docs :(
    GENERIC_GROUP = "GenericGroup"
    GENERIC_GROUP_PAGE = "GenericGroupPage"
    HTML_GROUP = "HTMLGroup"
    HTML_PAGE = "HTMLPage"
    ROOM = "Room"
    SCHEDULE = "Schedule"
    SCHEDULE_DAY = "ScheduleDay"
    SESSION = "Session"
    SESSION_GROUP = "SessionGroup"
    SESSION_SEGMENT = "SessionSegment"  # Spelt SessionSegement in API docs :(
    SESSION_TYPE = "SessionType"
    SPEAKER = "Speaker"
    SPEAKERS = "Speakers"
    SPONSOR = "Sponsor"
    SPONSORS = "Sponsors"
    STREAM = "Stream"


class Icon(IntEnum):
    """
    All icons can be found here: https://situ.entegy.com.au/Docs/Api/document-icons.
    """

    FACEBOOK = 1
    LINKEDIN = 2
    TWITTER = 3
    MAPPIN = 4
    GLOBE = 5
    ENEVELOPE = 6
    PHONE = 7
    SPEAKER = 8
    FILMREEL = 9
    DESKTOP = 10
    CHAT = 11
    MAP = 12
    ARTICLE = 13
    GALLERY = 14
    OUTBOX = 15
    LINK = 16
    PIN = 17
    BARGRAPH = 18
    PAPERCLIP = 19
    ACCESSPOINT = 20
    CALENDAR = 21
    CHECKMARK = 22
    PAPERPLANE = 23
    STAR = 24
    WARNINGTRIANGLE = 25
    SEARCH = 26
    FOLDER = 27
    INBOX = 28
    EDIT = 31
    PEOPLE = 32
    DOCUMENT = 35
    HOME = 37
    HOMESEARCH = 38
    DINER = 39
    OPENSIGN = 40
    INFOCIRCLE = 41
    HELPCIRCLE = 42
    FLAG = 43
    WINEGLASS = 44
    COCKTAIL = 45
    COFFEE = 46
    NEWS = 47
    CAR = 48
    OFFICEBUILDING = 49
    INFOSQUARE = 52
    GLOBESEARCH = 53
    ALARMCLOCK = 54
    GUITAR = 55
    ROADSIGNS = 56
    WIFITABLET = 57
    OPENBOOK = 58
    XBOX = 62
    GEAR = 64
    LOGOUT = 65
    INSTAGRAM = 68
    GOOGLEPLUS = 69
    HASHTAG = 70
    BACK = 71
    FORWARD = 72
    REFRESH = 73
    DOCUMENTSTAMP = 75
    ALARMBELL = 77
    ALARMBELLCOG = 78


class Document(BaseModel):
    name: str
    externalReference: str
    icon: Icon
    fileUrl: str


class ExternalContent(BaseModel):
    name: str
    externalReference: str
    icon: Icon
    fileUrl: str
    type: str


class Link(BaseModel):
    templateType: TemplateType
    moduleId: Optional[int] = None  # Optional when creating, not optional when retrieving.
    externalReference: str


class NamedLink(Link):
    name: str


class Category(BaseModel):
    externalReference: str
    name: str
    moduleId: Optional[int] = None  # Optional when creating, not optional when retrieving.
    childCategories: list["Category"] = []


class Content(BaseModel):
    name: str
    templateType: TemplateType
    externalReference: str
    mainImage: Optional[str] = None
    strings: dict[str, str]
    contentType: Optional[str] = None
    moduleId: Optional[int] = None  # Optional when creating, not optional when retrieving.
    pageSettings: Optional[dict[PageSetting, bool]] = None
    sortOrder: Optional[int] = None
    documents: Optional[list[Document]] = None
    links: Optional[list[Link]] = None
    multiLinks: Optional[list[NamedLink]] = None
    selectedCategories: Optional[list[Category]] = None
    children: Optional[list[Content]] = None

    def get_updated_content(self, new_content: Content) -> dict[str, Any]:
        """
        Get a dictionary of updated content when comparing another content object against self.
        The list of items that can be updated is guided by Entegy's API docs however a number of the
        optional attributes have been left in. Time will tell if this is a valid assumption.

        Return should really be an object. See:
        https://github.com/SituDevelopment/entegy-sdk-python/issues/184

        Parameters
        ----------
            `other` (`Content`): The content object against which to compare self.

        Returns
        -------
            `dict[str, Any]`: The results dict. Could really use an object here. To be fixed in
            https://github.com/SituDevelopment/entegy-sdk-python/issues/184
        """
        old_content = self  # for improved comparison readability.
        updated_content = {}

        # mainImage is a special case as Entegy mutates what we send them.
        # This means that every subsequent comparison will be falsy which would then kick off an update.
        if not old_content.mainImage and new_content.mainImage:
            updated_content["mainImage"] = new_content.mainImage

        if old_content.name != new_content.name:
            updated_content["name"] = new_content.name
        if old_content.strings != new_content.strings:
            updated_content["strings"] = new_content.strings
        if old_content.pageSettings != new_content.pageSettings:
            updated_content["pageSettings"] = new_content.pageSettings
        if old_content.sortOrder != new_content.sortOrder:
            updated_content["sortOrder"] = new_content.sortOrder
        if old_content.documents != new_content.documents:
            updated_content["documents"] = new_content.documents
        if old_content.links != new_content.links:
            updated_content["links"] = new_content.links
        if old_content.multiLinks != new_content.multiLinks:
            updated_content["multiLinks"] = new_content.multiLinks
        if old_content.selectedCategories != new_content.selectedCategories:
            updated_content["selectedCategories"] = new_content.selectedCategories

        # The update of content requires the presence of externalReference even though
        # it may not be one of the fields that have been changed. :(
        # Adding it here just in case.
        updated_content["externalReference"] = old_content.externalReference

        return updated_content


class ContentCreate(BaseModel):
    name: str
    templateType: TemplateType
    externalReference: str


class ContentChildCreate(BaseModel):
    name: str
    externalReference: Optional[str] = None
    mainImage: Optional[str] = None
    strings: Optional[dict[str, str]] = None
    links: Optional[list[Link]] = None
